(function () {
  const root = this;

  function isLink(field) {
    if (field) {

      return field.start && field.end;
    }
  }

  function isNode(field) {
    return !isLink(field);
  }

  /*
   * In future we may need neo4j driver object for the integer comparison and conversion.
   * That API is lacking though, so I write some of my own number functions in this file for now.
   * https://github.com/neo4j/neo4j-javascript-driver#a-note-on-numbers-and-the-integer-type
   * Current requirement is for neo4j object to be passed in (neo4j-driver v1.1)
   */
  function Parser(neo4j, ids) {
    ids = ids || { nodeIdField: 'identity', linkIdField: 'name' };
    this.neo4j = neo4j;
    this.nodeIdField = ids.nodeIdField || 'id';
    this.linkIdField = ids.linkIdField;
  }

  Parser.prototype.parse = function (boltResponse) {
    this.nodeDict = {};

    this.items = [];

    boltResponse.records.forEach(record => {
      this.parseFields(record._fields);
    })
    return this.items;
  };

  /*
   * Create the node and push to the items array if it does not yet exist
   */
  Parser.prototype.makeNode = function (field) {
    const id = `node${field.identity}`;
    const props = this.convertNumberProps(field.properties);
    if (!this.items[id]) {
      this.items.push({
        type: 'node',
        id,
        t: field.properties.title || field.properties.name,
        c: 'grey',
        d: {
          props: props,
          labels: field.labels
        }
      });
      this.nodeDict[id] = true;
    }
    return id;
  };

  Parser.prototype.makeLink = function (field, id1, id2) {
    const id = `link${field.identity}`;
    const props = this.convertNumberProps(field.properties);
    this.items.push({
      type: 'link',
      id: id,
      id1: id1,
      id2: id2,
      d: {
        props: props,
        labels: field.labels
      }
    })
  };

  Parser.prototype.convertNumberProps = function (props) {
    const updatedProps = [];
    Object.keys(props).forEach(key => {
      const prop = props[key];
      if (this.neo4j.isInt(prop)) {
        updatedProps[key] = {
          raw: prop,
          converted: this.convertInt(prop)
        };
      }
    });
    return updatedProps;
  };

  Parser.prototype.convertInt = function (neoInt) {
    return this.neo4j.integer.inSafeRange(neoInt) ? this.neo4j.integer.toNumber(neoInt) : neoInt;
  };

  // Beware: IDs/identities in neo4j are unique to their type (so a node and link could have the same ID)
  Parser.prototype.parseFields = function (fields) {
    const neoIdDict = {};
    fields.forEach(field => {
      const neoId = (isNode(field) ? 'node' : 'link') + field.identity.toString();
      neoIdDict[neoId] = isNode(field) ? this.makeNode(field) : field;
    })
    // now we have valid node IDs and a dictionary, we can parse the links
    Object.keys(neoIdDict).forEach(key => {
      const field = neoIdDict[key];
      if (isLink(field)) {
        const start = 'node' + field.start.toString();
        const end = 'node' + field.end.toString();
        this.makeLink(field, start, end);
      }
    });
  };

  const klbolt = {
    Parser: Parser
  };

  if (typeof module !== 'undefined' && module.exports) {
    module.exports = klbolt;
  } else {
    root.klbolt = klbolt;
  }

})();
